#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Mar  8 11:30:41 2021

@author: Heikki Välisuo
"""

import random
import paramsPandemia as pars
from Tilasto import tilasto
from Kartta import kartta
import Fantasmi

''' txtbeg
<h1> Pöpö fantasiamaailmassa </h1>

<p>Fantasiamaailmaan — yhteen monista vaihtoehtoisista todellisuuksista —
livahtaa ilkeä pöpö, jonka aikaansaannoksia seuraamme.
Onneksi tämä fantasiamaailma on tietokoneeni muistissa ja siksi pöpökin
on tosielämän elioille vaaraton.</p>
<hr />

<p style='margin-top: 2.0em'>Fantasmi on jossain seuraavista tiloista: </p>
<ul>
<li>altis: Terve, mutta altis tartunnalle</li>
<li>oireeton: Saanut tartunnan, mutta sitä ei voi havaita edes
testaamalla</li>
<li>sairas: Oireilee pikkuisen, mutta pyörii porukoissa.
Liikkuu terveitä vaisummin. Voi tartuttaa lähelle tulevia.</li>
<li>karanteenissa: Ei saa liikkua, ei tavata muita. Ei tartuta</li>
<li>kuollut: Jotkut kuolevat</li>
<li>immuuni: Tervehtymisen jälkeen potilas on pitkän aikaa immuuni.
</li>
<li>rokotettu: rokotettu voi sairastua, mutta ei niin helposti
kuin muut.</li>
<ul>
<p>Fantasmi kuuluu yhteen kansoista A, B, C tai D</p>
<hr />
<p style='margin-top: 2.0em'>
Luodaan listallinen fantasmeja. Pythonin lista on luettelo eli
<code>vaesto = [fantasmi_0, fantasmi_1, ... fantasmi_n]</code>
</p>
txtend '''

vaesto = [Fantasmi.Fantasmi() for i in range(pars.alku_vaesto)]

''' txtbeg
<p>Pääohjelma, joka pitää fantasiamaailman oliot — fantasmit —
liikkeessä</p>
txtend '''

def simulate():
    ''' txtbeg
    <p>Pandemian etenemisessä on paljon satunnaisuutta.
    <samp>random.seed()</samp> antaa joka simulaatioon erilaiset satunnaisluvut
    — ei tule kahta samanlaista simulaatiota.
    </p>
    txtend '''
    random.seed()

    ''' txtbeg
    <p>Tämän tunkeilijan mukana lintukotoon pääsee hypähtämään pöpö.</p>
    txtend '''

    tunkeilija()

    # Yksi while-silmukan kierros vastaa yhtä fantasiamaailman päivää
    # Simuloidaan kunnes ei tapahdu mitään uuttaa jännää
    while not (kartta.loppu or Fantasmi.ohjaus.loppu):

        # Tarkistetaan, tuleeko uusia tartuntoja
        # poimitaan väestöstä tartunnalle alttiit fantasmit ...
        alttiit = [fantasmi for fantasmi in vaesto
                   if fantasmi.tila in ['altis', 'rokotettu']]

        # ... ja potentiaaliset tartuttajat
        tartuttajat = [fantasmi for fantasmi in vaesto if
                       fantasmi.tila in ['oireeton', 'sairas']]

        # Ovatko jotkut yllälöydetyistä vaarallisessa lähikontaktissa?
        altistukset(tartuttajat, alttiit)

        # Tarkistetaan fantasmi fantasmilta tapahtuuko kenenkään
        # terveydentilassa muutoksia
        for fantasmi in vaesto:
            fantasmi.taudin_kulku()
            # ja siirretään fantasmia eteenpäin
            fantasmi.liike()

        # Päivitetään fantasmien sijainti komentokeskuksen näytöllä
        kartta.sijainnit(vaesto)
        # Päivitetään fantasiamaailman terveystilastot
        tilasto.aika_kuluu(vaesto)

        # Jos komentokeskus on käynnistänyt rokotuskampanjan,
        # rokotetaan sen verran kuin kyetään.
        if Fantasmi.ohjaus.rokota:
            rokotukset(vaesto)

        # Tarkistetaan komentokeskuksen strategiasta,
        # mitä tehdä huomenna
        Fantasmi.ohjaus.uusi_vaihe()
        # <<<< Uusi päivä. while-silmukaan alkuun.

    # Simulointi on päättynyt.
    # Kunnioitetaan hetki pandemian uhreja.
    kartta.hautausmaa(vaesto)

    print('päivä: ', tilasto.paiva)
    kartta.lopeta()


""" txtbeg
<p>Tarkistetaan, saako joku alttiista tartunnan joltakin sairaista</p>
txtend """

def altistukset(tartuttajat, alttiit):
    for tartuttaja in tartuttajat:
        for altis in alttiit:
            tartunta(tartuttaja, altis)

""" txtbeg
<p><samp>altis</samp> sairastuu, jos on joutunut liian lähelle
viruksen kantajaa — <samp>tartuttajaa</samp>.
<p><samp>altis</samp> on tästä eteenpäin tilassa 'oireeton'
<samp>kesto</samp> päivää.
Rokotettu parantuu puolessa ajassa.</p>
txtend """

def tartunta(tartuttaja, altis):
    if lahella(tartuttaja, altis):
        altis.tila = 'oireeton'
        altis.uusiTartunta = 3
        (t0, t1) = pars.kestot['oireeton']
        kesto = random.uniform(t0, t1)
        if altis.tila == 'rokotettu':
            kesto = kesto/2.0
        altis.ohi = tilasto.paiva + kesto
        tilasto.N_tartunnat += 1

    ''' txtbeg
    <p>Tarkistetaan, ovatko <samp>tartuttaja</samp> ja <samp>altis</samp>
    tartuntaetäisyydellä. Sovelletaan Pythagoraan lausetta</p>
    txtend '''

def lahella(tartuttaja, altis):
    suoja_etaisyys_2 = pars.suoja_etaisyys
    if altis.tila == 'rokotettu':
        suoja_etaisyys_2 = pars.suoja_etaisyys_vac
    # else:
    #     print('mitä hittoa ', tartuttaja.tila, altis.tila)
    (ax, ay) = tartuttaja.paikka
    (bx, by) = altis.paikka
    return (ax-bx)**2 + (ay-by)**2 < suoja_etaisyys_2

""" txtbeg
<p>Päivässä rokotetaan <samp>pars.rokotustahti</samp> verran
fantasmeja.</p>
<p>Itsellenikin yllätykseksi ensin rokotetaan fantasmi-kansa A,
sitten B, C ja viimeiseksi D. Selitys:</p>
<p>Rokotukset oli helpoin ohjelmoida alkaen fantasmien listan alusta.
Kartalle piirtämistä varten fantasmien lista taas kannattaa
laittaa kansojen mukaiseen järjestykseen.</p>
txtend """

def rokotukset(vaesto):
    i = 0
    for fantasmi in [x for x in vaesto if x.tila in ('altis', 'immuuni')]:
        fantasmi.tila = 'rokotettu'
        tilasto.N_rokotetut += 1
        i += 1
        if i > pars.rokotustahti:
            break

''' txtbeg
<p>Tällainen on se alussa mainittu väestön sekaan pujahtava tunkeilija.
Tunkeilijan tauti on oireettomassa itämisvaiheessa. Oireeton siirtyy tilaan
'sairas', kun on kulunut satunnainen aika väliltä (t0, t1)
</p>
txtend '''

def tunkeilija():
    # Luodaan yksi fantasmi
    tautinen = Fantasmi.Fantasmi()
    tautinen.koti = 'A'
    # Haluan pandemian alkavan kansan 'A' joukosta
    # Tämän voi kommentoida, jos haluaa satunnaisen alkukohdan
    tautinen.paikka = (0.3*pars.x_max, 0.7*pars.y_max)
    # Itämisvaiheessa 'lintukotoon tullessa
    tautinen.tila = 'oireeton'
    (t0, t1) = pars.kestot['oireeton']
    # Oireet tulevat esiin satunnaisen ajan kuluttua
    tautinen.ohi = tilasto.paiva + random.uniform(t0, t1)
    # Lisätään väestön sekaan
    vaesto.append(tautinen)
    # Tilastoluvut kasvavat yhdellä
    tilasto.N_alku += 1
    tilasto.N_vaesto += 1

# Tästä se lähtee käyntiin. Pidä hatustasi kiinni
simulate()
